// regFind.cpp
//
// this file enumerates registry keys and values
//

#include "Main.h"

// these are defined in WinNT.h
//
/*
 REG_NONE                    ( 0 )   // No value type
 REG_SZ                      ( 1 )   // Unicode nul terminated string
 REG_EXPAND_SZ               ( 2 )   // Unicode nul terminated string
                                            // (with environment variable references)
 REG_BINARY                  ( 3 )   // Free form binary
 REG_DWORD                   ( 4 )   // 32-bit number
 REG_DWORD_LITTLE_ENDIAN     ( 4 )   // 32-bit number (same as REG_DWORD)
 REG_DWORD_BIG_ENDIAN        ( 5 )   // 32-bit number
 REG_LINK                    ( 6 )   // Symbolic Link (unicode)
 REG_MULTI_SZ                ( 7 )   // Multiple Unicode strings
 REG_RESOURCE_LIST           ( 8 )   // Resource list in the resource map
 REG_FULL_RESOURCE_DESCRIPTOR ( 9 )  // Resource list in the hardware description
 REG_RESOURCE_REQUIREMENTS_LIST ( 10 )
 REG_QWORD                   ( 11 )  // 64-bit number
 REG_QWORD_LITTLE_ENDIAN     ( 11 )  // 64-bit number (same as REG_QWORD)
//*/

// used to track the entire tree branch for a key
char gRegKeyName[65536];
DWORD glNumberOfDiscrepancies = 0;

// prints the reg node tree
void printRegNodeTree(FILE *fptr, REG_KEY_NODE *pRegKey, int level)
{
	REG_DATA_NODE *pRegData;
	static char space[1024];
	static int lastLevelPrinted = -1;
	bool flag;

	if(pRegKey == NULL) return;

	memset(space, ' ', level * 2);
	space[level * 2] = 0;
	while(pRegKey != NULL)
	{
		if(lastLevelPrinted > level) fprintf(fptr, "\n");
		fprintf(fptr, "%s[%3u]  %s @ %s\n", space, level, pRegKey->regKeyName, pRegKey->dateTime);	// print the key info
		lastLevelPrinted = level;
		pRegData = pRegKey->regDataNode;

		flag = false;	// only print extra space if there is data
		while(pRegData != NULL)
		{
			flag = true;
			fprintf(fptr, "%s+--> [%3u]  (%02u) (%05u) %s\n", space, level, pRegData->dataType, pRegData->dataLength, pRegData->regDataName);
			pRegData = pRegData->nextData;
		}
		if(flag) fprintf(fptr, "\n");

		if(pRegKey->subKey != NULL)
		{
			printRegNodeTree(fptr, pRegKey->subKey, level+1);
			space[level * 2] = 0;
		}
		pRegKey = pRegKey->nextKey;
	}
	return;
} // printRegNodeTree

// prints the reg node sub tree, does not print the "nextKey" for initial node
void printRegNodeSubTree(FILE *fptr, REG_KEY_NODE *pRegKey, int level)
{
	REG_DATA_NODE *pRegDataNode;
	char space[256];

	if(pRegKey == NULL) return;

	memset(space, ' ', level * 2);
	space[level * 2] = 0;

	fprintf(fptr, "%s[%3u]  %s @ %s\n", space, level, pRegKey->regKeyName, pRegKey->dateTime);

	pRegDataNode = pRegKey->regDataNode;
	while(pRegDataNode != NULL)
	{
		fprintf(fptr, "%s+--> [%3u]  (%02u) (%05u) %s\n", space, level, pRegDataNode->dataType, pRegDataNode->dataLength, pRegDataNode->regDataName);
		pRegDataNode = pRegDataNode->nextData;
	}
	fprintf(fptr, "\n");

	if(pRegKey->subKey != NULL)
	{
		printRegNodeTree(fptr, pRegKey->subKey, level+1);
		fprintf(fptr, "\n");
		space[level * 2] = 0;
	}

	return;
} // printRegNodeSubTree

// allocate memory for a new registry node
REG_KEY_NODE *newRegKeyNode(char *name, int nameLen, char *dateTime)
{
	REG_KEY_NODE *pRegKey;

	pRegKey = (REG_KEY_NODE *) malloc( sizeof(REG_KEY_NODE) );
	if(pRegKey == NULL)
	{
		fprintf(stderr, "\n\nFatal Memory Error! Could not allocate %u bytes for a REGNODE.\n\n", sizeof(REG_KEY_NODE) );
		exit(-1);
	}

	pRegKey->regKeyName = (char *) malloc(nameLen + 1);
	if(pRegKey->regKeyName == NULL)
	{
		fprintf(stderr, "\n\nFatal Memory Error! Could not allocate %d bytes for regKeyName.\n\n", nameLen + 1 );
		exit(-1);
	}

	// copy name into node
	strcpy(pRegKey->regKeyName, name);

	memset(pRegKey->dateTime, 0, 24);
	if(dateTime != NULL)
		strncpy(pRegKey->dateTime, dateTime, 19);
	else
		pRegKey->dateTime[0] = 0;

	pRegKey->alreadyMatched = false;

	pRegKey->nextKey = NULL;
	pRegKey->subKey = NULL;
	pRegKey->regDataNode = NULL;

	return(pRegKey);
} // newRegKeyNode

// allocate memory for a new registry node
REG_DATA_NODE *newRegDataNode(char *name, int nameLen, DWORD dataType, DWORD dataLen)
{
	REG_DATA_NODE *pRegData;

	pRegData = (REG_DATA_NODE *) malloc( sizeof(REG_DATA_NODE) );
	if(pRegData == NULL)
	{
		fprintf(stderr, "\n\nFatal Memory Error! Could not allocate %u bytes for a REG_DATA_NODE.\n\n", sizeof(REG_DATA_NODE));
		exit(-1);
	}

	pRegData->regDataName = (char *) malloc(nameLen + 1);
	if(pRegData->regDataName == NULL)
	{
		fprintf(stderr, "\n\nFatal Memory Error! Could not allocate %d bytes for regDataName.\n\n", nameLen + 1 );
		exit(-1);
	}

	strcpy(pRegData->regDataName, name);
	pRegData->dataType = dataType;
	pRegData->dataLength = dataLen;
	pRegData->pdata = NULL;

	pRegData->nextData = NULL;

	return(pRegData);
} // newRegDataNode

// checks the date and time differential
void checkDateTime(FILE *fptr, char *dateTime1, char *dateTime2, char *path, char *keyName) // 2005:08:31 20:17:55
{
	int x;

	x = strnicmp(dateTime1, dateTime2, 19);
	if(x == 0) return; // they match

	//	Key data/time different
	fprintf(fptr, "%3d) <7> Date/time for Key Name: \"%s\"\n\n     in Key: \"%s\"\n\n     -- changed from: %s to date/time: %s\n\n", 
		++glNumberOfDiscrepancies, keyName, path, dateTime1, dateTime2);

	return;
} // checkDateTime

// finds a matching key
REG_KEY_NODE *findMatchingKeys(REG_KEY_NODE *pRegKey1, REG_KEY_NODE *pRegKey2, REG_KEY_NODE *pEndRegKey2)
{
	while(pRegKey2 != NULL)
	{
		if(stricmp(pRegKey1->regKeyName, pRegKey2->regKeyName) == 0) return(pRegKey2);	// found a match

		pRegKey2 = pRegKey2->nextKey;

		if(pRegKey2 == pEndRegKey2) return NULL;	// optimization, gives a stopping point
	}
	return pRegKey2;
} // findMatchingKeys

void printData(FILE *fptr, BYTE *data, long length)
{
	long x, max;

	max = length;
	if(max > 256) max = 256;

	fprintf(fptr, "        Printed Length (%u):\n", max);
	fprintf(fptr, "          ");
	for(x = 1; x <= max; x++)
	{
		fprintf(fptr, "%02x ", data[x-1]);
		if(x % 16 == 0) fprintf(fptr, "\n          ");
	}
	fprintf(fptr, "\n\n\n");

	return;
} // printData

void compareKeyData(FILE *fptr, long dataLength1, long dataLength2, char *keyPath1, char *keyPath2)
{
	static long index1 = 0, index2 = 0;
	int x;

	if(index1 >= gFileSize1)
	{
		fprintf(stderr, "Index1 (%d) has exceeded file size1 (%d)\n\n", index1, gFileSize1);
		return;
	}

	if(index2 >= gFileSize2)
	{
		fprintf(stderr, "Index2 (%d) has exceeded file size2 (%d)\n\n", index2, gFileSize2);
		return;
	}

	if(dataLength1 == dataLength2)	// could be a match
	{
		x = memcmp( &gpRegData1[index1], &gpRegData2[index2], dataLength1 );
		if( x == 0 )
		{
			index1 += dataLength1;
			index2 += dataLength2;
			return;	// match
		}

		fprintf(fptr, "%3d) <B> Data from Key1: \"%s\"\n\n      not equal to Data from Key2: \"%s\"\n\n", ++glNumberOfDiscrepancies, keyPath1, keyPath2);

		fprintf(fptr, "      Data from File1 - Total Length:%u\n\n", dataLength1);
		printData(fptr, &gpRegData1[index1], dataLength1);

		fprintf(fptr, "      Data from File2 - Total Length:%u\n\n", dataLength2);
		printData(fptr, &gpRegData2[index2], dataLength2);

		index1 += dataLength1;
		index2 += dataLength2;
		return;
	}

	fprintf(fptr, "%3d) <C> Data from Key1: \"%s\"\n\n      different size than Key2: \"%s\"\n\n", ++glNumberOfDiscrepancies, keyPath1, keyPath2);

	fprintf(fptr, "      Data from File1:\n\n");
	printData(fptr, &gpRegData1[index1], dataLength1);

	fprintf(fptr, "      Data from File2:\n\n");
	printData(fptr, &gpRegData2[index2], dataLength2);

	index1 += dataLength1;
	index2 += dataLength2;
	return;
} // compareKeyData

// checks two lists of data nodes for descrepancies
void compareDataNodes(FILE *fptr, REG_DATA_NODE *pRegDN1, REG_DATA_NODE *pRegDN2, char *keyPath1, char *keyPath2, char *logFile1, char *logFile2)
{
	REG_DATA_NODE *tmpRegDN;

	if(pRegDN1 == NULL && pRegDN2 == NULL) return;	//match, neither have data values

	// 2nd reg has data values which do not exist in 1st reg
	if(pRegDN1 == NULL && pRegDN2 != NULL)
	{
		fprintf(fptr, "%3d) <5> Data Nodes below do not exist in file1: \"%s\"\n", ++glNumberOfDiscrepancies, logFile1);
		while(pRegDN2 != NULL)
		{
			fprintf(fptr, "          ----: \"%s\"\n", pRegDN2->regDataName);
			pRegDN2 = pRegDN2->nextData;
		}
		fprintf(fptr, "\n");
		return;
	}

	// 1st reg has data values which do not exist in 2nd reg
	if(pRegDN1 != NULL && pRegDN2 == NULL)
	{
		fprintf(fptr, "%3d) <6> Data Nodes below do not exist in file2: \"%s\"\n", ++glNumberOfDiscrepancies, logFile2);
		while(pRegDN1 != NULL)
		{
			fprintf(fptr, "          ----: \"%s\"\n", pRegDN1->regDataName);
			pRegDN1 = pRegDN1->nextData;
		}
		fprintf(fptr, "\n");
		return;
	}

	// both have data nodes, let's see how well they match
	while(pRegDN1 != NULL)
	{
		tmpRegDN = pRegDN2;
		while(tmpRegDN != NULL)	// linear search, generally, these lists will be fairly short
		{
			if( stricmp(pRegDN1->regDataName, tmpRegDN->regDataName) == 0) break;	// found a match

			tmpRegDN = tmpRegDN->nextData;
		}

		if(tmpRegDN == NULL)	// this data value not found
		{
			fprintf(fptr, "%3d) <8> Data Value: \"%s\" not found in file2: \"%s\" \n\n", ++glNumberOfDiscrepancies, pRegDN1->regDataName, logFile2);
			pRegDN1 = pRegDN1->nextData;
			continue;
		}

		// found a data node match, check the data type
		if(pRegDN1->dataType != tmpRegDN->dataType)
		{
			fprintf(fptr, "%3d) <9> Data Type (%u) for Data Name: \"%s\" in key1: \"%s\"\n\n     --- different than\n\n         Data Type (%u) in key2: \"%s\"\n\n", 
				++glNumberOfDiscrepancies, pRegDN1->dataType, pRegDN1->regDataName, keyPath1, tmpRegDN->dataType, keyPath2);
		}

		// check the data length
		if(pRegDN1->dataLength != tmpRegDN->dataLength)
		{
			fprintf(fptr, "%3d) <A> Data Length (%u) for Data Name: \"%s\" in key1: \"%s\"\n\n     --- different than\n\n         Data Length (%u) in key2: \"%s\"\n\n", 
				++glNumberOfDiscrepancies, pRegDN1->dataLength, pRegDN1->regDataName, keyPath1, tmpRegDN->dataLength, keyPath2);
		}

		//check the data itself
		if(gSaveRegistryData)	// jao check data
		{
			compareKeyData(fptr, pRegDN1->dataLength, tmpRegDN->dataLength, keyPath1, keyPath2);
		}

		pRegDN1 = pRegDN1->nextData;
	} // end while loop

	return;
} // compareDataNodes

// compares two registry logs to look for differences
DWORD match2Registries(FILE *fptr, REG_KEY_NODE *pTopRegKey1, REG_KEY_NODE *pTopRegKey2, int level, char *keyPath1, char *keyPath2, char *logFile1, char *logFile2)
{
	REG_KEY_NODE *pRegKey1, *pRegKey2, *pTmpRegKey2;

	int p1, p2;

	// loop and check key values
	pRegKey1 = pTopRegKey1;
	pRegKey2 = pTopRegKey2;
	while(pRegKey1 != NULL)
	{
		// find a matching key
		pTmpRegKey2 = findMatchingKeys(pRegKey1, pRegKey2, NULL);	// start at this point, NULL means go to end of list
		if(pTmpRegKey2 == NULL)	// didn't find from this point forward, check from start
		{
			pTmpRegKey2 = findMatchingKeys(pRegKey1, pTopRegKey2, pRegKey2);	// start at this point, NULL means go to end of list
			if(pTmpRegKey2 == NULL)	// didn't find anywhere on the list
			{
				fprintf(fptr, "%3d) <1> Key1: \"%s\"\n\n     Name: \"%s\"\n\n     -- not found in file2: \"%s\"\n\n",  ++glNumberOfDiscrepancies,
					keyPath1, pRegKey1->regKeyName, logFile2);

				pRegKey1 = pRegKey1->nextKey;
				continue;
			}
		}

		// found a matching key at this level
		pTmpRegKey2->alreadyMatched = true;	// set this so we don't revisit when we iterate registry log file 2
		compareDataNodes(fptr, pRegKey1->regDataNode, pTmpRegKey2->regDataNode, keyPath1, keyPath2, logFile1, logFile2);
		checkDateTime(fptr, pRegKey1->dateTime, pTmpRegKey2->dateTime, keyPath1, pRegKey1->regKeyName);

		// check sub keys
		if(pRegKey1->subKey == NULL && pTmpRegKey2->subKey == NULL)	// match, neither keys have a sub-key
		{
			pRegKey1 = pRegKey1->nextKey;
			continue;
		}

		// if file2 has a subkey, but file1 does not
		if(pRegKey1->subKey == NULL && pTmpRegKey2->subKey != NULL)
		{
			fprintf(fptr, "%3d) <2> Key2: \"%s\"\n\n     -- missing subkey: \"%s\" in file1: \"%s\"\n\n", ++glNumberOfDiscrepancies,
				keyPath1, pTmpRegKey2->subKey->regKeyName, logFile1);
			printRegNodeSubTree(fptr, pTmpRegKey2, level);

			pRegKey1 = pRegKey1->nextKey;
			continue;
		}

		// if file1 has a subkey, but file2 does not
		if(pRegKey1->subKey != NULL && pTmpRegKey2->subKey == NULL)
		{
			fprintf(fptr, "%3d) <3> Key2: \"%s\"\n\n     -- missing subkey: \"%s\" in file2: \"%s\"\n\n",  ++glNumberOfDiscrepancies,
				keyPath2, pRegKey1->subKey->regKeyName, logFile2);
			printRegNodeSubTree(fptr, pRegKey1, level);

			pRegKey1 = pRegKey1->nextKey;
			continue;
		}

		// they both have subkeys
		p1 = (int) strlen(keyPath1);
		p2 = (int) strlen(keyPath2);
		strcat(keyPath1, "\\");
		strcat(keyPath2, "\\");
		strcat(keyPath1, pRegKey1->regKeyName);
		strcat(keyPath2, pTmpRegKey2->regKeyName);

		match2Registries(fptr, pRegKey1->subKey, pTmpRegKey2->subKey, level+1, keyPath1, keyPath2, logFile1, logFile2);

		keyPath1[p1] = 0;
		keyPath2[p2] = 0;

		pRegKey1 = pRegKey1->nextKey;
	} // end while looping through reg1 data values

	// check for any key data values in list 2 not found in list one
	pRegKey2= pTopRegKey2;
	while(pRegKey2 != NULL)
	{
		if(pRegKey2->alreadyMatched)		// this IS a key OR data value already matched
		{
			pRegKey2 = pRegKey2->nextKey;
			continue;
		}

		pRegKey2->alreadyMatched = true;
		fprintf(fptr, "%3d) <4> Key2: \"%s\"\n\n     Name: \"%s\"\n\n     -- not found in file1: \"%s\"\n\n", 
			++glNumberOfDiscrepancies, keyPath2, pRegKey2->regKeyName, logFile1);

		pRegKey2 = pRegKey2->nextKey;
	} // end while looping through list 2

	return glNumberOfDiscrepancies;
} // match2Registries

// find node to attach new node to
REG_KEY_NODE *findRegNamePos(REG_KEY_NODE *pTop, REG_KEY_NODE *pCur, char *keyName)	// finds the node to add to next
{
	REG_KEY_NODE *pReg, *prevNode;
	int x;

	// search the list from the current position
	// in tests with two 24 MB registry log files, this optimization shaved 6 of 9 seconds consistently!
	pReg = pCur;
	prevNode = NULL;
	while(pReg != NULL)	// start here because likelihood is high it is on the list after last entry
	{
		x = (int) stricmp(pReg->regKeyName, keyName);	// should not be equal

		if(x > 0)	// pReg->regKeyName > keyName
		{
			if(prevNode == NULL)	// this name is before cur, will have to search from begining
				break;		
			else
				return(prevNode);	// already have a prev node
		}
		prevNode = pReg;
		pReg = pReg->nextKey;
	}
	if(pReg == NULL) return(prevNode);	// put on end of list

	pReg = pTop;		// start at top of list this time
	prevNode = NULL;
	while(pReg != NULL)
	{
		x = (int) stricmp(pReg->regKeyName, keyName);	// should not be equal

		if(x > 0) return(prevNode);	// pReg->regKeyName > keyName
		prevNode = pReg;
		pReg = pReg->nextKey;
	}

	return(prevNode);
} // findRegNamePos

// read the registry log file
// NOTE: first line already read
REG_KEY_NODE *readRegFile(FILE *fptr)
{
	int keyLevel, numSpace, startPos, nameLen;
	static char line[4096];
	char keyName[1024], *tmp;
	REG_KEY_NODE *tmpRegKeyNode, *tmpRN;
	REG_DATA_NODE *tmpRegDataNode, *curRDN;

	REG_KEY_NODE *pTopmostRegKeyNode, *pTopRegKeyNode[MAX_REG_LEVEL], *pCurRegKeyNode[MAX_REG_LEVEL];
	DWORD dataType, dataLength;

	pTopmostRegKeyNode = NULL;
	for(keyLevel = 0; keyLevel < MAX_REG_LEVEL; keyLevel++ )		// clear all node pointers
	{
		pTopRegKeyNode[keyLevel] = NULL;
		pCurRegKeyNode[keyLevel] = NULL;
	}

	// one root node
	pTopmostRegKeyNode = newRegKeyNode("ROOT_NODE", 9, NULL);
	pTopRegKeyNode[0] = pTopmostRegKeyNode;
	pCurRegKeyNode[0] = pTopmostRegKeyNode;

	keyLevel = 0;
	while(1)
	{
		tmp = fgets(line, 4096, fptr);
		if(tmp == NULL) return(pTopmostRegKeyNode);	// DONE
		if(line[0] == 0x0a || line[0] == 0x0d) continue;		// skip blank lines

		if(line[0] == 'L') continue;	// skip lines with level information

		// get the number of blank spaces
		numSpace = 0;
		while(line[numSpace] == ' ') numSpace++;

		// process the key or value
		if(line[numSpace] == '"')
		{
			startPos = numSpace + 1;	// should be on a double quote, plus 1 is first character of name
			numSpace = (numSpace / 2) + 1; // now equals level that this key is expected to be on
			if(numSpace >= MAX_REG_LEVEL)
			{
				fprintf(stderr, "\n\nFatal Error! Exceeded maximum number of levels (%u) - likey error in the file.\n\n", MAX_REG_LEVEL);
				exit(-1);
			}

			nameLen = 0;
			while(line[startPos + nameLen] != '"') nameLen++;

			// copy the key name into a temp variable
			memcpy(keyName, &line[startPos], nameLen);
			keyName[nameLen] = 0;	// NULL terminate

			tmpRN = newRegKeyNode(keyName, nameLen, &line[startPos + nameLen + 2]);

			// add to list at level indicated by (spaces/2+1), this covers when level is the same, or increases
			if(numSpace <= keyLevel)	// level indicated equals current level, will ALWAYS be at least one existing node
			{
				// finds the node to add to next
				keyLevel = numSpace;
				tmpRegKeyNode = findRegNamePos(pTopRegKeyNode[keyLevel], pCurRegKeyNode[keyLevel], keyName);

				// if tmpRegKeyNode is NULL, that means it goes at top of list, before current first node
				if(tmpRegKeyNode == NULL)
				{
					tmpRN->nextKey = pTopRegKeyNode[keyLevel];
					pTopRegKeyNode[keyLevel] = tmpRN;
					pCurRegKeyNode[keyLevel-1]->subKey = tmpRN;
				}
				else
				{
					tmpRN->nextKey = tmpRegKeyNode->nextKey;
					tmpRegKeyNode->nextKey = tmpRN;
				}
				pCurRegKeyNode[keyLevel] = tmpRN;	// keep this to help find insertion point more quickly, 
										    	// since Registry generally in order, AND for when coming up a level
			}

			// go down a level (should only be one level lower)
			else if(numSpace > keyLevel)
			{
				if(pCurRegKeyNode[keyLevel]->subKey != NULL)	// no other nodes on lower level
				{
					fprintf(stderr, "\n\nError! Trying to add additional subkeys to existing key list.\n\n");
					printRegNodeTree(stderr, pTopmostRegKeyNode, 0);
					return(NULL);
				}

				pCurRegKeyNode[keyLevel]->subKey = tmpRN;
				keyLevel++;
				pTopRegKeyNode[keyLevel] = tmpRN;
				pCurRegKeyNode[keyLevel] = tmpRN;
			}
		}	// end if key value, i.e. srch for "

		// check for data value
		else if(line[numSpace] == '(')
		{
			startPos = numSpace + 1;	// should be on an open parenthesis, plus 1 is first character of name
			numSpace = (numSpace / 2) + 1; // now equals level that this data value SHOULD be on
			if(numSpace >= MAX_REG_LEVEL)
			{
				fprintf(stderr, "\n\nFatal Error! Exceeded maximum number of levels (%u) - likey error in the file.\n\n", MAX_REG_LEVEL);
				exit(-1);
			}

			dataType = atol(&line[startPos]);	// get data type
			dataLength = atol(&line[startPos + 5]);

			startPos += 13;	// points to first letter in data value name
			nameLen = 0;
			while(nameLen < 1024)	// should never reach 1024 (limit of keyname) or 4096
			{
				if(line[startPos + nameLen] == '"')	// some key value names have quotes in the name
					if(line[startPos + nameLen + 1] == 0x0d || line[startPos + nameLen + 1] == 0x0a) break;	// quote followed by end of line
	
				nameLen++; // find ending quote
			}

			// copy the data value name into a temp variable
			memcpy(keyName, &line[startPos], nameLen);
			keyName[nameLen] = 0;	// NULL terminate

			tmpRegDataNode = newRegDataNode(keyName, nameLen, dataType, dataLength);
			curRDN = pCurRegKeyNode[numSpace-1]->regDataNode;	// num spaces can never be zero

			if(curRDN == NULL)	// nohting on the data node list, point to this one
			{
				pCurRegKeyNode[numSpace-1]->regDataNode = tmpRegDataNode;
			}
			else
			{
				while(curRDN->nextData != NULL)	// find last element - this will be a short list
				{
					curRDN = curRDN->nextData;
				}
				curRDN->nextData = tmpRegDataNode;
			}
		} // end if line begins with open parenthesis
		else
		{
			fprintf(stderr, "\n\nError, neither quote nor parenthesis, found as first character=%x on line=%s (len=%u)\n\n", line[0], line, strlen(line));
		}
	} // end while loop through reg log file

	return(pTopmostRegKeyNode);
} // readRegFile

// search through all registry keys
void recurseRegKeys(HKEY hKey)
{
	static int level = 0, lastLevel = 0;
	static char space[2048] = { 0 };
	static char subKeyName[2048];		// had some stack limitations for deep diving
	static char valueName[16384];

	DWORD sizeName, indx, keyNameLen, spaceLen;
	LONG result;
	FILETIME fileTime;
	SYSTEMTIME sysFileTime;
	HKEY subKey;

	strcat(space, "  ");
	spaceLen = (DWORD) strlen(space);
	level++;	// increase the level

	// loop through sub keys
	indx = 0;
	while(1)
	{
		sizeName = 2048;	// reset max size
		result = RegEnumKeyEx(hKey, indx, subKeyName, &sizeName, NULL, NULL, NULL, &fileTime);

if(strcmp(subKeyName, "Wow6432") == 0) { printf("WOW 64 32\n"); exit(0); }

		if(result == ERROR_SUCCESS)
		{
			if( level != lastLevel )
			{
				lastLevel = level;
				fprintf(gFptr, "\nLk%02u:%s\n", level, gRegKeyName);	// level of the key
			}

			memset(&sysFileTime, 0, sizeof(SYSTEMTIME));
			FileTimeToSystemTime(&fileTime, &sysFileTime);	// GMT time

			fprintf(gFptr, "%s\"%s\" %04d:%02d:%02d %02d:%02d:%02d\n", space, subKeyName, 
				sysFileTime.wYear, sysFileTime.wMonth, sysFileTime.wDay, 
				sysFileTime.wHour, sysFileTime.wMinute, sysFileTime.wSecond);

			// try to open any sub keys of this key
			subKey = hKey;
			result = RegOpenKeyEx(subKey, subKeyName, 0, KEY_READ | CROSS_ACCESS, &subKey);		// KEY_READ | KEY_WOW64_64KEY, KEY_WOW64_32KEY

			if(result == ERROR_SUCCESS) // if there is a sub key, recurse
			{
				strcat(gRegKeyName, "\\");	// ~
				strcat(gRegKeyName, subKeyName);

				// enumerate subkey
				recurseRegKeys(subKey);
				RegCloseKey(subKey);

				// remove the added subkey from the path name
				keyNameLen = strlen(gRegKeyName);
				while(keyNameLen > 0)
				{
					if(gRegKeyName[keyNameLen] != '\\')	// ~
					{
						gRegKeyName[keyNameLen] = 0;
						keyNameLen -= 1;
						continue;
					}

					gRegKeyName[keyNameLen] = 0;
					break;
				}
			} // end if open key successful
		} // end if enum key success
		else if(result == ERROR_NO_MORE_ITEMS)
		{
			break;
		}
		else
		{
			fprintf(stderr, "(%d) Error (%d), getting subkey.\n", indx, result);
			fprintf(stderr, "(%d) subKeyName = %s, size=%u, timelo=%u, timeHi=%u\n\n", indx,
				subKeyName, sizeName, fileTime.dwLowDateTime, fileTime.dwHighDateTime);
		}
		indx++;
	} // end while looping through sub-keys

	// enumerate the values under a key
	indx = 0;
	int flag = 0;

	// enumerate the values of this sub-key
	while(1)
	{
		DWORD dataType, dataLength, tmpLen;
		BYTE dataBuffer[1024];	// large enough for most keys, avg data is 29 bytes
		BYTE *ptrDataBuffer;

		sizeName = 16384;
		result = RegEnumValue(hKey, indx, valueName, &sizeName, NULL, &dataType, NULL, &dataLength);	// not getting data

if(strcmp(valueName, "Wow6432") == 0) { printf("2.WOW 64 32\n"); exit(0); }

		if(result == ERROR_SUCCESS)	// want ERROR_MORE_DATA here because will be valid if getting data and size exceeds 1024
		{
			if(valueName[0] != 0)
			{
				if(flag == 0)	// first time for this value
				{
					flag = -1;
					fprintf(gFptr, "Lv:%02u\n", level);	// level of the value
				}
				fprintf(gFptr, "%s(%02u) (%05u) \"%s\"\n", space, dataType, dataLength, valueName);
			}
			else if(dataLength != 0)	// may have "(Default)" as name, but still have a data value
			{
				if(flag == 0)	// first time for this value
				{
					flag = -1;
					fprintf(gFptr, "Lv:%02u\n", level);	// level of the value
				}
				fprintf(gFptr, "%s(%02u) (%05u) \"(Default)\"\n", space, dataType, dataLength);
			}

			// retrieve the data if successfuly got other info
			// data size is unlimited - for purpose of comparing and knowing the old data, we can limit data size
			// in registry, largest data size has been ... ~150K
			// retrieve the data
			if(gSaveRegistryData)	// jao save reg data
			{
				ptrDataBuffer = dataBuffer;
				if(dataLength > 1024)	// didn't get all the data, try again
				{
					ptrDataBuffer = (BYTE *) malloc(dataLength);
					if(ptrDataBuffer == NULL)// unlikley to happen
					{
						fprintf(stderr, "\n\nWarning - Could not allocate %u bytes for registry data - Data File is now corrupt, delete it.\n\n", dataLength);
						dataLength = 0;	// prevents reading data and freeing of NULL buffer
					}
				}

				sizeName = 16384;	// need to reset for 2nd call or will get the ERROR_MORE_DATA msg
				result = RegEnumValue(hKey, indx, valueName, &sizeName, NULL, &dataType, ptrDataBuffer, &dataLength);	// get the values for the key

				// if buffer was not allocated, then this write will not occur
				if(result == ERROR_SUCCESS)
				{
					fwrite(ptrDataBuffer, 1, dataLength, gfptrRegData1);
				}
				if(dataLength > 1024) free(ptrDataBuffer);	// free it now, this is a recursive routine, optimization: static and only reallocate if too small
			}

		} // end result==ERROR_SUCCESS
		else if(result == ERROR_NO_MORE_ITEMS)
		{
			break;
		}
		else
		{
			fprintf(stderr, "(%d) Error (%d), getting value.\n", indx, result);
			fprintf(stderr, "(%d) valueName = %s, size=%u\n\n", indx, valueName, sizeName);
		}

		indx++;
	} // end while looping through key values

	if(spaceLen >= 2)
		space[spaceLen-2] = 0;	// remove added space characters
	else
		space[0] = 0;

	level--;
	return;
} // recurseRegKeys

void searchRegistry(SYSTEMTIME *sysTime)
{
	HKEY hKey;
	LONG result;
	char logFileName[MAX_PATH], data[] = "Data", dataFileName[MAX_PATH];

	if(!gSaveRegistryData)	// if not a registry data list
	{
		data[0] = 0;
	}
	sprintf(logFileName, "RegList%s_%04d_%02d_%02d_%02d_%02d_%02d.log", data, 
		sysTime->wYear, sysTime->wMonth, sysTime->wDay, 
		sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

	// open the file
	gFptr = fopen(logFileName, "w+");
	if(gFptr == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open log file %s for writing.\n\n", logFileName);
		return;
	}

	if(gSaveRegistryData)	// give a name easily reconstructed when comparing files
	{
		sprintf(dataFileName, "RegList%s_%04d_%02d_%02d_%02d_%02d_%02d.log_DF.dat", data, 
			sysTime->wYear, sysTime->wMonth, sysTime->wDay, 
			sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

		gfptrRegData1 = fopen(dataFileName, "wb+");
		if(gfptrRegData1 == NULL)
		{
			fprintf(stderr, "\n\nWarning - Could not open data file %s for writing - will turn it off.\n\n", dataFileName);
			gSaveRegistryData = false;	// turn it off
		}
	}

	fprintf(stdout, "\nSearching the registry\n\n");

	if(gSaveRegistryData)	// check again because it could get turned off
	{
		fprintf(gFptr, "REGISTRY+DATA\n");
	}
	else
	{
		fprintf(gFptr, "REGISTRY\n");
	}

	result = RegOpenKeyEx(HKEY_CLASSES_ROOT, NULL, 0, KEY_READ | CROSS_ACCESS, &hKey);
	if(result == ERROR_SUCCESS)
	{
		fprintf(stdout, "\n\nHKEY_CLASSES_ROOT\n\n", hKey);
		fprintf(gFptr, "\n\nL00\n\"HKEY_CLASSES_ROOT\"\n\n", hKey);
		memset(gRegKeyName, 0, sizeof(gRegKeyName) );
		strcpy(gRegKeyName, _T("HKEY_CLASSES_ROOT") );
		recurseRegKeys(hKey);
		RegCloseKey(hKey);
	}

	result = RegOpenKeyEx(HKEY_CURRENT_USER, NULL, 0, KEY_READ | CROSS_ACCESS, &hKey);
	if(result == ERROR_SUCCESS)
	{
		fprintf(stdout, "\n\nHKEY_CURRENT_USER\n\n", hKey);
		fprintf(gFptr, "\n\nL00\n\"HKEY_CURRENT_USER\"\n\n", hKey);
		memset(gRegKeyName, 0, sizeof(gRegKeyName) );
		strcpy(gRegKeyName, _T("HKEY_CURRENT_USER") );
		recurseRegKeys(hKey);
		RegCloseKey(hKey);
	}

	result = RegOpenKeyEx(HKEY_LOCAL_MACHINE, NULL, 0, KEY_READ | CROSS_ACCESS, &hKey);
	if(result == ERROR_SUCCESS)
	{
		fprintf(stdout, "\n\nHKEY_LOCAL_MACHINE\n\n", hKey);
		fprintf(gFptr, "\n\nL00\n\"HKEY_LOCAL_MACHINE\"\n\n", hKey);
		memset(gRegKeyName, 0, sizeof(gRegKeyName) );
		strcpy(gRegKeyName, _T("HKEY_LOCAL_MACHINE") );
		recurseRegKeys(hKey);
		RegCloseKey(hKey);
	}

	result = RegOpenKeyEx(HKEY_USERS, NULL, 0, KEY_READ | CROSS_ACCESS, &hKey);
	if(result == ERROR_SUCCESS)
	{
		fprintf(stdout, "\n\nHKEY_USERS\n\n", hKey);
		fprintf(gFptr, "\n\nL00\n\"HKEY_USERS\"\n\n", hKey);
		memset(gRegKeyName, 0, sizeof(gRegKeyName) );
		strcpy(gRegKeyName, _T("HKEY_USERS") );
		recurseRegKeys(hKey);
		RegCloseKey(hKey);
	}

	result = RegOpenKeyEx(HKEY_PERFORMANCE_DATA, NULL, 0, KEY_READ | CROSS_ACCESS, &hKey);
	if(result == ERROR_SUCCESS)
	{
		fprintf(stdout, "\n\nHKEY_PERFORMANCE_DATA\n\n", hKey);
		fprintf(gFptr, "\n\nL00\n\"HKEY_PERFORMANCE_DATA\"\n\n", hKey);
		memset(gRegKeyName, 0, sizeof(gRegKeyName) );
		strcpy(gRegKeyName, _T("HKEY_PERFORMANCE_DATA") );
		recurseRegKeys(hKey);
		RegCloseKey(hKey);
	}

	result = RegOpenKeyEx(HKEY_CURRENT_CONFIG, NULL, 0, KEY_READ | CROSS_ACCESS, &hKey);
	if(result == ERROR_SUCCESS)
	{
		fprintf(stdout, "\n\nHKEY_CURRENT_CONFIG\n\n", hKey);
		fprintf(gFptr, "\n\nL00\n\"HKEY_CURRENT_CONFIG\"\n\n", hKey);
		memset(gRegKeyName, 0, sizeof(gRegKeyName) );
		strcpy(gRegKeyName, _T("HKEY_CURRENT_CONFIG") );
		recurseRegKeys(hKey);
		RegCloseKey(hKey);
	}

	result = RegOpenKeyEx(HKEY_DYN_DATA, NULL, 0, KEY_READ | CROSS_ACCESS, &hKey);
	if(result == ERROR_SUCCESS)
	{
		fprintf(stdout, "\n\nHKEY_DYN_DATA\n\n", hKey);
		fprintf(gFptr, "\n\nL00\n\"HKEY_DYN_DATA\"\n\n", hKey);
		memset(gRegKeyName, 0, sizeof(gRegKeyName) );
		strcpy(gRegKeyName, _T("HKEY_DYN_DATA") );
		recurseRegKeys(hKey);
		RegCloseKey(hKey);
	}

	if(gSaveRegistryData) fclose(gfptrRegData1);
	fclose(gFptr);
	return;
} // searchRegistry

